import collections.abc
import numbers

import numpy as np

from slxpy.common.mapping import dtype_mapping, suffix_mapping

def is_nonstr_sequence(x):
    """Check if x is a sequence, but not a string."""
    return isinstance(x, collections.abc.Sequence) and not isinstance(x, str)

def format_integer(x, dtype: np.dtype):
    iinfo = np.iinfo(dtype)
    if iinfo.min <= x <= iinfo.max:
        return str(x)
    else:
        raise ValueError(f"Integer value {x} out of range for dtype {dtype}")

def format_float(x, dtype: np.dtype):
    ctype = dtype_mapping[dtype.name]
    special = {
        "inf": f"std::numeric_limits<{ctype}>::infinity()",
        "-inf": f"-std::numeric_limits<{ctype}>::infinity()",
    }
    if isinstance(x, str):
        return special[x]
    else:
        suffix = suffix_mapping[dtype.name]
        return f"{x}{suffix}"

def format_number(x, dtype: np.dtype):
    if issubclass(dtype.type, numbers.Integral):
        return format_integer(x, dtype)
    elif issubclass(dtype.type, numbers.Real):
        return format_float(x, dtype)
    else:
        raise ValueError(f"Unexpected dtype: {dtype}")

def format_sequence(x, dtype: np.dtype):
    return f"{{ {', '.join(format_number(y, dtype) for y in x)} }}"

def format_sequence_or_scalar(x, dtype: np.dtype):
    return format_sequence(x, dtype) if is_nonstr_sequence(x) else format_number(x, dtype)



from typing import List
import numpy as np

def tsort(am: np.ndarray) -> List[int]:
    assert am.shape[0] == am.shape[1], f"Adjacency matrix must be square, but got shape {am.shape}"
    verts = am.shape[0]

    l = []  # List that will contain the sorted elements
    s = np.zeros(verts, dtype=np.bool_)
    s[~am.any(axis=0)] = 1

    # Kahn's algorithm
    while True:
        nz = s.nonzero()[0]
        if nz.size == 0: break
        vert = nz[0]
        s[vert] = 0
        l.append(vert)
        row: np.ndarray = am[vert, :]
        indices = row.nonzero()[0]
        row.fill(0)
        col: np.ndarray = am[:, indices]
        s[indices[~col.any(axis=0)]] = 1

    if np.any(am):
        raise ValueError("Adjacency matrix contains circular structure.")
    else:
        return l



from collections.abc import Sequence
from dataclasses import dataclass, field, fields
from functools import cached_property
from slxpy.common.enum import FieldMode
from typing import Any, Callable, ClassVar, Dict, List, Literal, Optional, TYPE_CHECKING, TextIO, Tuple, Union
import numpy as np
import json

import slxpy.common.constants as C
from slxpy.common.tsort import tsort
from slxpy.common.env_config import Config as EnvConfig

# NOTE: Currently nested struct hierarchy is dropped, seeking use-case and better implementation.

@dataclass
class Field:
    name: str
    doc: str

    mode: FieldMode
    raw_shape: Optional[List[int]]  # Available when mode is *_ARRAY
    type: Optional[str]  # Available when mode is STRUCT or STRUCT_ARRAY

    @property
    def size(self):
        prod = 1
        if self.shape is not None:
            for el in self.shape:
                prod *= el
        return prod

    @property
    def shape(self):
        """
        NOTE: temporary patch for MATLAB limitation for row-major layout
        https://ww2.mathworks.cn/help/rtw/ug/code-generation-of-matrix-data-and-arrays.html
        In brief, lack of support for some fundemental blocks such as Integrator and Transfer Fcn,
        which is awkward and annoying.
        So, revert `UseRowMajorAlgorithm` to `off` and `ArrayLayout` to `Column-major`
        Transpose dimensions here to make a F-contiguous array C-contiguous.
        See also: tune_codegen_config.m
        """
        if self.raw_shape is None: return None
        if len(self.raw_shape) == 2 and (self.raw_shape[0] == 1 or self.raw_shape[1] == 1):
            # Workaround for 1-dimensional InstP array got generated as [x 1]
            # Consider put the logic here or MATLAB script postprocess.m
            # Check whether it is safe
            return [self.raw_shape[0] * self.raw_shape[1]]
        else:
            return self.raw_shape[::-1]

    @staticmethod
    def reconstruct(d: dict):
        mode = FieldMode(d["mode"])
        if mode == FieldMode.PLAIN or mode == FieldMode.MODEL:
            shape, type = None, None
        elif mode == FieldMode.PLAIN_ARRAY:
            shape, type = d["shape"], None
        elif mode == FieldMode.STRUCT:
            shape, type = None, d["type"]
        elif mode == FieldMode.STRUCT_ARRAY:
            shape, type = d["shape"], d["type"]
        elif mode == FieldMode.ENUM:
            shape, type = None, d["type"]
        elif mode == FieldMode.ENUM_ARRAY:
            shape, type = d["shape"], d["type"]
        elif mode == FieldMode.POINTER:
            shape, type = None, d["type"]
        elif mode == FieldMode.POINTER_ARRAY:
            shape, type = d["shape"], d["type"]
        else:
            raise ValueError(f"Unsupported field mode: {mode}")

        return Field(
            name=d["name"],
            doc=d.get("doc", ""),
            mode=mode,
            raw_shape=shape,
            type=type
        )

    def asdict(self, dict_filter):
        d = {
            "name": self.name,
            "doc": self.doc,
            "mode": self.mode.value,
            "shape": self.raw_shape,
            "type": self.type
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

@dataclass
class Enumerator:
    name: str
    doc: str
    value: int

    @staticmethod
    def reconstruct(d: dict):
        return Enumerator(name=d["name"], doc=d.get("doc", ""), value=d["value"])

    def asdict(self, dict_filter):
        d = { "name": self.name, "doc": self.doc, "value": self.value }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

@dataclass
class Type:
    name: str
    doc: str
    location: Literal["root", "model_class"]
    is_enum: bool
    fields: Union[List[Field], List[Enumerator]]  # NOTE: Maybe bad design

    _id: int = field(init=False)
    _alias_name: str = field(init=False)
    _binding_name: str = field(init=False)
    _binding_identifier: str = field(init=False)

    def __post_init__(self):
        self._binding_name = self.name
        self._binding_identifier = f"{self._binding_name}_PB"

    def _assign_index(self, index: int):
        assert index >= 0
        self._id = index
        self._alias_name = f"SlxpyPodType_{index}"

    @cached_property
    def field_dict(self):
        return { field.name: field for field in self.fields }

    @staticmethod
    def reconstruct(d: dict):
        name = d["name"]
        location = d["location"]
        is_enum = d["is_enum"]
        FieldCls = Enumerator if is_enum else Field
        t = Type(
            name=name,
            doc=d.get("doc", ""),
            location=location,
            is_enum=is_enum,
            fields=[FieldCls.reconstruct(f) for f in d["fields"]]
        )
        return t

    def asdict(self, dict_filter):
        d = {
            "name": self.name,
            "doc": self.doc,
            "location": self.location,
            "is_enum": self.is_enum,
            "fields": [f.asdict(dict_filter) for f in self.fields],
            "_id": self._id,
            "_alias_name": self._alias_name,
            "_binding_name": self._binding_name,
            "_binding_identifier": self._binding_identifier
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

class TypeContainer(Sequence):
    def __init__(self, types: List[Type], model_class: 'ModelClass'):
        super().__init__()
        self._mc = model_class
        self._storage: List[Type] = []
        self._name_map: Dict[str, int] = {}
        for type in types: self._add(type)

    def _add(self, type: Type):
        if type.location == "model_class":
            identifier = f"{self._mc.identifier}::{type.name}"
        elif type.location == "root":
            identifier = type.name
        else:
            raise ValueError("Unexpected struct location.")
        index = len(self._storage)
        self._name_map[identifier] = index
        type._assign_index(index)
        self._storage.append(type)

    def lookup(self, name: str):
        for type in self._storage:
            if type.name == name: return type
        raise LookupError("Can't find type")

    @cached_property
    def dependency_order(self):
        return self.topo_sorted("dependency")

    @cached_property
    def hierarchy_order(self):
        return self.topo_sorted("hierarchy")

    @cached_property
    def in_root(self):
        return [type for type in self._storage if type.location == "root"]

    @cached_property
    def in_model_class(self):
        return [type for type in self._storage if type.location == "model_class"]

    def topo_sorted(self, order: Literal["dependency", "hierarchy"]):
        if order == "dependency":
            am = self._dependency_adjacency_matrix()
        elif order == "hierarchy":
            am = self._hierarchy_adjacency_matrix()
        else:
            raise ValueError("Unexpected sort order")
        indices = tsort(am)
        sorted_list = []
        for i in indices:
            sorted_list.append(self._storage[i])
        return sorted_list

    def _dependency_adjacency_matrix(self):
        verts = len(self._storage)
        am = np.zeros((verts, verts), dtype=np.bool_)
        for col, item in enumerate(self._storage):
            for field in item.fields:
                if isinstance(field, Field) and field.type is not None:
                    if field.type == "void": continue
                    row = self._name_map[field.type]
                    am[row, col] = 1
        return am

    def _hierarchy_adjacency_matrix(self):
        raise NotImplementedError("Currently nested struct hierarchy is dropped")

    def __getitem__(self, index):
        return self._storage[index]

    def __len__(self):
        return len(self._storage)

    @staticmethod
    def reconstruct(ds: list, model_class: 'ModelClass'):
        return TypeContainer([Type.reconstruct(d) for d in ds], model_class)

    def asdicts(self, dict_filter):
        return [t.asdict(dict_filter) for t in self._storage]

@dataclass
class Method:
    name: str
    doc: str
    # parameters, policy

    @staticmethod
    def reconstruct(d: dict):
        return Method(name=d["name"], doc=d.get("doc", ""))

    def asdict(self, dict_filter):
        d = { "name": self.name, "doc": self.doc }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

@dataclass
class ModelClass:
    name: str
    doc: str
    namespace: str
    identifier: str
    sample_time: float
    methods: List[Method]
    fields: List[Field]
    field_mapping: Dict[str, str]
    type_mapping: Dict[str, str]

    _alias_name: str = "SlxpyExtensionModelClass"
    _binding_name: str = field(init=False)
    _binding_identifier: str = field(init=False)

    def __post_init__(self):
        self._binding_name = self.name
        self._binding_identifier = f"{self._binding_name}_PB"

    @staticmethod
    def reconstruct(d: dict):
        return ModelClass(
            name=d["name"],
            doc=d.get("doc", ""),
            namespace=d.get("namespace", ""),
            identifier=d["identifier"],
            sample_time=d["sample_time"],
            methods=[Method.reconstruct(f) for f in d["methods"]],
            fields=[Field.reconstruct(f) for f in d["fields"]],
            field_mapping=d["field_mapping"],
            type_mapping=d["type_mapping"],
        )

    def asdict(self, dict_filter):
        d = {
            "name": self.name,
            "doc": self.doc,
            "namespace": self.namespace,
            "identifier": self.identifier,
            "sample_time": self.sample_time,
            "methods": [m.asdict(dict_filter) for m in self.methods],
            "fields": [f.asdict(dict_filter) for f in self.fields],
            "field_mapping": self.field_mapping,
            "type_mapping": self.type_mapping,
            "_alias_name": self._alias_name,
            "_binding_name": self._binding_name,
            "_binding_identifier": self._binding_identifier
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

@dataclass
class Module:
    name: str
    doc: str
    version: str
    author: str
    license: str
    model_class: ModelClass
    types: TypeContainer
    env: EnvConfig

    @staticmethod
    def reconstruct(d: dict):
        model_class = ModelClass.reconstruct(d["model_class"])
        types = TypeContainer.reconstruct(d["types"], model_class)
        return Module(
            name=d["name"],
            doc=d.get("doc", ""),
            version=d["version"],
            author=d["author"],
            license=d["license"],
            model_class=model_class,
            types=types,
            env=EnvConfig.reconstruct(d["env"])
        )

    def asdict(self, dict_filter):
        d = {
            "name": self.name,
            "doc": self.doc,
            "version": self.version,
            "author": self.author,
            "license": self.license,
            "model_class": self.model_class.asdict(dict_filter),
            "types": self.types.asdicts(dict_filter),
            "env": self.env.asdict(dict_filter)
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

@dataclass
class Context:
    sources: List[str]
    headers: List[str]
    module: Module
    VERSION: ClassVar[str] = C.metadata_version

    @staticmethod
    def load(fp: TextIO):
        d = json.load(fp)
        assert d["__version__"] == Context.VERSION, "Context version incompatible"
        return Context.reconstruct(d)

    def dump(self, fp: TextIO, debug: bool = False):
        if debug:
            dict_filter = lambda x:x
        else:
            def _concise_predicate(pair: Tuple[str, Any]):
                k, v = pair
                return (
                    not k.startswith("_")
                    and v is not None
                    and (k != "doc" or v != "")
                )
            dict_filter: Callable[[dict], dict] = lambda x: {k:v for k, v in x.items() if _concise_predicate((k, v))}
        d = {"__version__": Context.VERSION, **self.asdict(dict_filter)}
        json.dump(d, fp, sort_keys=False)

    @staticmethod
    def reconstruct(d: dict):
        return Context(
            sources=d["sources"],
            headers=d["headers"],
            module=Module.reconstruct(d["module"])
        )

    def asdict(self, dict_filter):
        d = {
            "sources": self.sources,
            "headers": self.headers,
            "module": self.module.asdict(dict_filter)
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    def refresh(self):
        return Context.reconstruct(self.asdict())



from pathlib import Path
from typing import List

import click

import slxpy.common.constants as C
from slxpy.cli.utils import ensure_slxpy_project


@click.command()
@click.option('--all', is_flag=True, help="Complete cleanup, keeping config files only.")
@click.pass_context
def clean(ctx: click.Context, all: bool):
    """
    Clean up working directory.
    """
    workdir: Path = ctx.obj["workdir"]
    ensure_slxpy_project(workdir)

    import shutil
    def rmdir(dir: Path):
        if dir.exists():
            assert dir.is_dir()
            shutil.rmtree(dir)
            click.echo(f"Remove folder: {dir.relative_to(workdir)}")
        else:
            click.echo(f"Skip folder:   {dir.relative_to(workdir)}")

    def rmfile(file: Path):
        if file.exists():
            assert file.is_file()
            file.unlink()
            click.echo(f"Remove file:   {file.relative_to(workdir)}")
        else:
            click.echo(f"Skip file:     {file.relative_to(workdir)}")

    rmdir(workdir / "build")
    rmdir(workdir / "include")

    from slxpy.backend.renderer import assets
    generated_files: List[str] = [asset.name for asset in assets] + [C.project_ir_name]
    for f in generated_files:
        rmfile(workdir / f)

    if all:
        rmdir(workdir / "model")
        rmfile(workdir / C.metadata_name)




simulink_mapping = [
    ('block_signals'             , 'B'         , 'Block signals of the system'),
    # ('external_inputs'           , 'ExtU'      , 'Block input data for root system'),
    # ('external_input_sizes'      , 'ExtUSize'  , 'Size of block input data for the root system (used when inputs are variable dimensions)'),
    # ('external_outputs'          , 'ExtY'      , 'Block output data for the root system'),
    # ('external_output_sizes'     , 'ExtYSize'  , 'Size of block output data for the root system'),
    ('external_inputs'           , 'U'         , 'Input data'),
    ('external_input_sizes'      , 'USize'     , 'Size of input data'),
    ('external_outputs'          , 'Y'         , 'Output data'),
    ('external_output_sizes'     , 'YSize'     , 'Size of output data'),
    ('parameters'                , 'P'         , 'Parameters for the system'),
    ('const_block_signals'       , 'ConstB'    , 'Block inputs and outputs that are constants'),
    ('machine_local_data'        , 'MachLocal' , 'Used by ERT S-function targets'),
    ('const_parameters'          , 'ConstP'    , 'Constant parameters in the system'),
    ('const_parameters_with_init', 'ConstInitP', 'Initialization data for constant parameters in the system'),
    ('discrete_states'           , 'DW'        , 'Block states in the system'),
    ('mass_matrix'               , 'MassMatrix', 'Used for physical modeling blocks'),
    ('zero_crossing_states'      , 'PrevZCX'   , 'Previous zero-crossing signal state'),
    ('continuous_states'         , 'X'         , 'Continuous states'),
    ('disabled_states'           , 'XDis'      , 'Status of an enabled subsystem'),
    ('state_derivatives'         , 'XDot'      , 'Derivatives of continuous states at each time step'),
    ('zero_crossing_signals'     , 'ZCV'       , 'Zero-crossing signals'),
    # ('default_parameters'        , 'DefaultP'  , 'Default parameters in the system'),
    # ('global_TID'                , 'GlobalTID' , 'Used for sample time for states in referenced models'),
    # ('invariant_signals'         , 'Invariant' , 'Invariant signals'),
    # ('n_stages'                  , 'NSTAGES'   , 'Solver macro'),
    # ('object'                    , 'Obj'       , 'Used by ERT C++ code generation to refer to referenced model objects'),
    # ('timing_bridge'             , 'TimingBrdg', 'Timing information stored in different data structures'),
    # ('shared_DSM'                , 'SharedDSM' , 'Shared local data stores, which are Data Store Memory blocks with Share across model instances selected'),
    ('instance_parameters'       , 'InstP'     , 'Parameter arguments for the system'),
]

dtype_mapping = {
    "uint8"  : "uint8_t" ,
    "int8"   : "int8_t"  ,
    "uint16" : "uint16_t",
    "int16"  : "int16_t" ,
    "uint32" : "uint32_t",
    "int32"  : "int32_t" ,
    "uint64" : "uint64_t",
    "int64"  : "int64_t" ,
    "float32": "float"   ,
    "float64": "double"  ,
}

suffix_mapping = {
    "float32": "F" ,
    "float64": "",
}



import textwrap, subprocess, os, json
from pathlib import Path
from typing import List

import click


def check_conda_installation():
    try:
        get_conda_executable()
    except KeyError:
        raise click.ClickException(textwrap.dedent("""\
        Conda package is not installed in current environment.
        To use this command, you need to check three things:
            1. Ensure anaconda/miniconda is installed in your system.
            2. This command is run in a conda environment.
        """).strip())


def get_conda_env_set():
    def _postprocess(env: str):
        env_path = Path(env)
        return env_path.name

    args = ["env", "list"]
    env_text_list: List[str] = execute_conda_command(args)["envs"]
    return set(_postprocess(env) for env in env_text_list)


def create_conda_env(env_name: str, packages: str):
    args = ["create", "--name", env_name, "--channel", "conda-forge", "--override-channels", "--yes", *packages]
    run_conda_command(args)


def remove_conda_env(env_name: str):
    args = ["env", "remove", "--name", env_name, "--yes"]
    run_conda_command(args)


def run_in_conda_env(env_name: str, cwd: str, args: List[str]):
    args = ["run", "--name", env_name, "--no-capture-output", "--live-stream", "--cwd", cwd, *args]
    run_conda_command(args)


def get_conda_executable():
    return os.environ["CONDA_EXE"]


def execute_conda_command(args: List[str], json_output: bool = True):
    args = [get_conda_executable()] + args
    if json_output:
        args.append("--json")
    text = subprocess.check_output(args).decode("utf-8")
    if json_output:
        return json.loads(text)
    else:
        return text


def run_conda_command(args: List[str]):
    args = [get_conda_executable()] + args
    cp = subprocess.run(args)
    cp.check_returncode()



import jinja2

def create_jinja_env(package_path: str):
    # Import namespace package is broken in Jinja2 < 3.0
    # See: https://github.com/pallets/jinja/issues/1097
    # To avoid obscure errors, we check the version early.
    _check_jinja_version()
    loader = jinja2.PackageLoader(
        package_name="slxpy.templates",
        package_path=package_path,
        encoding="utf-8"
    )
    env = jinja2.Environment(
        # extensions=['jinja2.ext.debug'],
        loader=loader,
        auto_reload=False,
        autoescape=False,
        keep_trailing_newline=True,
        # trim_blocks=True,
    )
    env.policies['json.dumps_kwargs'] = {'sort_keys': True, 'ensure_ascii': False}
    return env

def _check_jinja_version():
    from packaging.version import parse as parse_version
    actual = parse_version(jinja2.__version__)
    required = parse_version('3.0')
    if actual < required:
        raise RuntimeError(
            f"jinja2 version {actual} is too old. "
            f"Please upgrade to {required} or later."
        )



import shutil, importlib_resources
from pathlib import Path
from typing import Callable, List
from dataclasses import dataclass
from slxpy.backend.filter import add_filters

from slxpy.common.context import Context, FieldMode
from slxpy.common.jinja import create_jinja_env
import slxpy.common.constants as C

@dataclass
class AssetInfo:
    name: str
    template: str
    condition: Callable[[Context], bool] = lambda _: True
    overwrite: bool = True

assets: List[AssetInfo] = [
    AssetInfo(
        name="module.cc",
        template="module.cc.jinja"
    ),
    AssetInfo(
        name="setup.py",
        template="setup.py.jinja"
    ),
    AssetInfo(
        name="pyproject.toml",
        template="pyproject.toml.jinja"
    ),
    AssetInfo(
        name="raw_env.h",
        template="raw_env.h.jinja",
        condition=lambda context: context.module.env.use_raw
    ),
    AssetInfo(
        name="common_env.h",
        template="common_env.h.jinja",
        condition=lambda context: context.module.env.use_raw or context.module.env.use_gym
    ),
    AssetInfo(
        name="gym_env.h",
        template="gym_env.h.jinja",
        condition=lambda context: context.module.env.use_gym
    ),
    AssetInfo(
        name="CMakeLists.txt",
        template="CMakeLists.txt.jinja"
    ),
    AssetInfo(
        name="test_extension.py",
        template="test_extension.py.jinja",
        overwrite=False
    )
]

includes = [
    "common.h",
    "bind.h",
    "complex.h",
    "data.h",
    "simulink_builtin.h",
    "env.h"
]


def render(workdir: Path, debug: bool = False):
    context_path = workdir / C.project_ir_name
    with context_path.open() as f: context = Context.load(f)

    if debug:
        context_debug_path = workdir / f"{context_path.stem}.debug.json"
        with context_debug_path.open("w") as f: context.dump(f, debug=True)

    env = create_jinja_env("backend")
    add_filters(env)
    env.globals["FieldMode"] = FieldMode

    active_assets = [asset for asset in assets if asset.condition(context)]
    optional_headers = [asset.name for asset in active_assets if asset.name.endswith(".h")]
    for asset in active_assets:
        asset_name, template = asset.name, asset.template
        asset_path = workdir / asset_name
        if asset.overwrite or not asset_path.exists():
            text = env.get_template(template).render(C=context, slxpy_headers=includes, optional_headers=optional_headers)
            asset_path.write_text(text)

    include_src = importlib_resources.files("slxpy.include")
    include_dst = workdir / "include"
    include_dst.mkdir(exist_ok=True)

    include_slxpy_dst = include_dst / "slxpy"
    include_slxpy_dst.mkdir(exist_ok=True)
    for inc in includes:
        inc_src = include_src / inc
        inc_dst = include_slxpy_dst / inc
        shutil.copyfile(inc_src, inc_dst)

    include_thirdparty = include_src / "third-party"
    shutil.copytree(include_thirdparty / "nlohmann", include_dst / "nlohmann", dirs_exist_ok=True)
    shutil.copytree(include_thirdparty / "fmt", include_dst / "fmt", dirs_exist_ok=True)



from dataclasses import dataclass, fields
from typing import ClassVar, Dict, List, Literal, Tuple, Type, Union
import numpy as np
from slxpy.common.formatter import format_sequence, format_sequence_or_scalar

from slxpy.common.mapping import dtype_mapping

@dataclass
class SpaceConfig:
    type: Literal["Box", "Discrete", "MultiDiscrete", "MultiBinary"]

    init_name: ClassVar[str] = ""
    mapping: ClassVar[Dict[str, Type['SpaceConfig']]] = {}
    def __init_subclass__(cls):
        SpaceConfig.mapping[cls.init_name] = cls

    @staticmethod
    def reconstruct(d: dict):
        type = d["type"]
        cls = SpaceConfig.mapping[type]
        return cls.reconstruct(d)

    def asdict(self, dict_filter):
        d = {
            "type": self.type
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @staticmethod
    def unique(inits: List['SpaceConfig']):
        return list(set(type(init) for init in inits))

    @staticmethod
    def default():
        return BoxSpaceConfig("Box", 0.0, 1.0, (2, 2), np.dtype("float64"))

@dataclass
class BoxSpaceConfig(SpaceConfig):
    low: Union[List[float], float]
    high: Union[List[float], float]
    shape: Tuple[int, ...]
    dtype: np.dtype
    init_name: ClassVar[str] = "Box"

    @staticmethod
    def reconstruct(d: dict):
        return BoxSpaceConfig(
            type=d["type"],
            low=d["low"],
            high=d["high"],
            shape=tuple(d["shape"]),
            dtype=np.dtype(d["dtype"])
        )

    def asdict(self, dict_filter):
        d = {
            "type": self.type,
            "low": self.low,
            "high": self.high,
            "shape": self.shape,
            "dtype": self.dtype.name
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @property
    def ctype(self):
        return dtype_mapping[self.dtype.name]

    @property
    def initializer(self):
        return f"{format_sequence_or_scalar(self.low, self.dtype)}, {format_sequence_or_scalar(self.high, self.dtype)}, {format_sequence(self.shape, np.dtype(np.int64))}"

    @property
    def func(self):
        return f"make_box<{dtype_mapping[self.dtype.name]}>"

@dataclass
class DiscreteSpaceConfig(SpaceConfig):
    n: int
    init_name: ClassVar[str] = "Discrete"
    func: ClassVar[str] = "make_discrete"
    ctype: ClassVar[str] = "int64_t"

    @staticmethod
    def reconstruct(d: dict):
        return DiscreteSpaceConfig(type=d["type"], n=d["n"])

    def asdict(self, dict_filter):
        d = { "type": self.type, "n": self.n }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @property
    def initializer(self):
        return f"{self.n}"

@dataclass
class MultiDiscreteSpaceConfig(SpaceConfig):
    nvec: List[int]
    init_name: ClassVar[str] = "MultiDiscrete"
    func: ClassVar[str] = "make_multi_discrete"
    ctype: ClassVar[str] = "int64_t"

    @staticmethod
    def reconstruct(d: dict):
        return MultiDiscreteSpaceConfig(type=d["type"], nvec=d["nvec"])

    def asdict(self, dict_filter):
        d = { "type": self.type, "nvec": self.nvec }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @property
    def initializer(self):
        return format_sequence(self.nvec)

@dataclass
class MultiBinarySpaceConfig(SpaceConfig):
    n: int
    init_name: ClassVar[str] = "MultiBinary"
    func: ClassVar[str] = "make_multi_binary"
    ctype: ClassVar[str] = "int8_t"

    @staticmethod
    def reconstruct(d: dict):
        return MultiBinarySpaceConfig(type=d["type"], n=d["n"])

    def asdict(self, dict_filter):
        d = { "type": self.type, "n": self.n }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @property
    def initializer(self):
        return f"{self.n}"



from dataclasses import dataclass, fields
from typing import ClassVar, Dict, List, Literal, TYPE_CHECKING, Tuple, Type, Union
import numpy as np

from slxpy.common.enum import FieldMode

if TYPE_CHECKING:
    from slxpy.common.context import Field

def check_shape(array_like, shape: List[int]):
    size = np.prod(shape).item()
    array = np.array(array_like)
    if array.size != 1:
        assert array.ndim == 1, "Currently only support one dimensional vector."
        assert array.size == size, f"Size inconsistency, broadcast is currently unsupported."

@dataclass
class InitConfig:
    type: Literal["seed", "constant", "uniform", "custom"]

    init_name: ClassVar[str] = ""
    mapping: ClassVar[Dict[str, Type['InitConfig']]] = {}
    def __init_subclass__(cls):
        InitConfig.mapping[cls.init_name] = cls

    @staticmethod
    def reconstruct(d: dict):
        type = d["type"]
        cls = InitConfig.mapping[type]
        if len(fields(cls)) > 1:
            return cls.reconstruct(d)
        else:
            return cls(type=type)

    def asdict(self, dict_filter):
        d = {
            "type": self.type
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @staticmethod
    def unique(inits: List['InitConfig']):
        return list(set(type(init) for init in inits))

    def check_basic_compatibility(self, field: "Field"):
        if field.mode in (FieldMode.STRUCT, FieldMode.STRUCT_ARRAY):
            raise ValueError("For struct parameters, use \"custom\" type initialization.")

@dataclass
class SeedInitConfig(InitConfig):
    init_name: ClassVar[str] = "seed"

@dataclass
class ConstantInitConfig(InitConfig):
    value: Union[float, str]
    init_name: ClassVar[str] = "constant"

    @staticmethod
    def reconstruct(d: dict):
        return ConstantInitConfig(
            type=d["type"],
            value=d["value"]
        )

    def asdict(self, dict_filter):
        d = {
            "type": self.type,
            "value": self.value
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

@dataclass
class UniformInitConfig(InitConfig):
    low: Union[float, List[float]]
    high: Union[float, List[float]]
    init_name: ClassVar[str] = "uniform"

    @staticmethod
    def reconstruct(d: dict):
        return UniformInitConfig(
            type=d["type"],
            low=d["low"],
            high=d["high"]
        )

    def asdict(self, dict_filter):
        d = {
            "type": self.type,
            "low": self.low,
            "high": self.high
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    def check_basic_compatibility(self, field: "Field"):
        super().check_basic_compatibility(field)
        low, high = self.low, self.high
        if field.mode == FieldMode.PLAIN_ARRAY:
            check_shape(low, field.shape)
            check_shape(high, field.shape)
        else:
            assert isinstance(low, float) and isinstance(high, float)

@dataclass
class CustomInitConfig(InitConfig):
    code: str
    init_name: ClassVar[str] = "custom"

    @staticmethod
    def reconstruct(d: dict):
        return CustomInitConfig(
            type=d["type"],
            code=d["code"]
        )

    def asdict(self, dict_filter):
        d = {
            "type": self.type,
            "code": self.code
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)



import sys
from pathlib import Path
from typing import Optional

import slxpy.common.constants as C


DEBUG = False
def set_debug(debug: bool):
    global DEBUG
    DEBUG = debug


def is_debug():
    return DEBUG


def ensure_slxpy_project(workdir: Path):
    # Safe check except init subcommand
    if not (workdir / C.model_config_name).exists() or not (workdir / C.env_config_name).exists():
        raise Exception("Not a slxpy project directory.")


def get_plat_name():
    import distutils.util
    return distutils.util.get_platform()


def get_plat_specifier(version: Optional[str] = None) -> str:
    # Ported from distutils/command/build.py
    if version is None:
        version = f"{sys.version_info.major}.{sys.version_info.minor}"
    plat_name = get_plat_name()
    return f".{plat_name}-{version}"



import re, textwrap, click
from pathlib import Path

from slxpy.common.jinja import create_jinja_env
from slxpy.common.model_config import Config as ModelConfig
from slxpy.common.env_config import Config as EnvConfig
from slxpy.common.fs import dir_empty
import slxpy.common.constants as C


def check_valid_workspace(workdir: Path):
    """
    Either raise or return
    """
    if workdir.exists():
        if workdir.is_dir():
            if not dir_empty(workdir):
                if (workdir / C.model_config_name).is_file():
                    raise ValueError(f"Folder {workdir} is already inited as a slxpy project. Choose another path instead.")
                else:
                    raise ValueError(f"Folder {workdir} has content. Expect an empty folder")
            # else: OK
        else:
            raise NotADirectoryError("Expect workdir to be a directory")
    # else: OK

def init_interactive(workdir: Path):
    check_valid_workspace(workdir)
    model, class_name, namespace = get_info_interactive(workdir)
    workdir.mkdir(parents=True, exist_ok=True)
    model_cfg_path, env_cfg_path = init(workdir, model, class_name, namespace)

    output_text = textwrap.dedent(f"""\
        Working directory: {workdir}
        Initialize config files:
            {model_cfg_path.name:<12} - Basic Simulink code generation config.
            {' ' * (12 + 3)}Adjust [simulink] and other sections as needed.
            {env_cfg_path.name:<12} - Raw/gym-like environment wrapper config.
    """).strip()
    click.echo(output_text)

def get_info_interactive(workdir: Path):
    workdir_name = workdir.name
    if workdir_name.isidentifier():
        model_default, class_name_default = workdir_name, f"{workdir_name}ModelClass"
    else:
        model_default, class_name_default = None, None

    matlab_naming = re.compile("^[a-zA-Z][a-zA-Z0-9_]*$")
    cpp_naming = re.compile("^[a-zA-Z_][a-zA-Z0-9_]*$")

    model: str = click.prompt("Simulink model name", type=str, default=model_default)
    assert matlab_naming.match(model), "Simulink model name is not valid."
    class_name: str = click.prompt("Code generation C++ class name", type=str, default=class_name_default)
    assert cpp_naming.match(class_name), "C++ class name is not valid."
    namespace: str = click.prompt("Code generation C++ namespace", type=str, default="")
    assert namespace == "" or cpp_naming.match(namespace), "C++ namespace name is not valid."

    return model, class_name, namespace

def init(workdir: Path, model: str, class_name: str, namespace: str):
    env = create_jinja_env('frontend')
    env.filters["tf"] = lambda x: "true" if x else "false"

    model_config = ModelConfig.default(model, class_name, namespace)
    model_cfg_text = env.get_template(f"{C.model_config_name}.jinja").render(config=model_config)
    model_cfg_path = workdir / C.model_config_name
    model_cfg_path.write_text(model_cfg_text)

    env_config = EnvConfig.default()
    env_cfg_text = env.get_template(f"{C.env_config_name}.jinja").render(config=env_config)
    env_cfg_path = workdir / C.env_config_name
    env_cfg_path.write_text(env_cfg_text)

    return model_cfg_path, env_cfg_path



from pathlib import Path

import click

from slxpy.cli.utils import is_debug, set_debug


@click.group()
@click.option('--debug', is_flag=True, help="Enable debug mode.")
@click.option('--workdir', '-w', default=".", type=click.Path(file_okay=False, resolve_path=True, path_type=Path), help="Working directory, must be empty or nonexistent. (Default to cwd)")
@click.pass_context
def app(ctx: click.Context, workdir: Path, debug: bool):
    set_debug(debug)

    ctx.ensure_object(dict)
    ctx.obj["workdir"] = workdir
    click.echo(f"Working directory: {workdir}")


def entry_point():
    try:
        app()
    except Exception as e:
        click.secho(str(e), fg="red", err=True)
        if is_debug():
            raise
        else:
            exit(1)


def _register_cli_commands(app: click.Group):
    from slxpy.cli.init import init
    from slxpy.cli.generate import frontend, backend, generate
    from slxpy.cli.multi_build import multi_build
    from slxpy.cli.clean import clean
    from slxpy.cli.pack import pack

    app.add_command(init)
    app.add_command(frontend)
    app.add_command(backend)
    app.add_command(generate)
    app.add_command(multi_build)
    app.add_command(clean)
    app.add_command(pack)


_register_cli_commands(app)



from enum import Enum

class FieldMode(Enum):
    MODEL = 0
    PLAIN = 1
    PLAIN_ARRAY = 2
    STRUCT = 3
    STRUCT_ARRAY = 4
    ENUM = 5
    ENUM_ARRAY = 6
    POINTER = 7
    POINTER_ARRAY = 8



import os, shutil
from pathlib import Path
from typing import Tuple


import click
from slxpy.cli.conda import check_conda_installation, create_conda_env, get_conda_env_set, remove_conda_env, run_in_conda_env

from slxpy.cli.utils import ensure_slxpy_project, get_plat_name, get_plat_specifier


@click.group()
@click.pass_context
def multi_build(ctx: click.Context):
    """
    Build project for multiple Python versions.
    """


@multi_build.command()
@click.option('--versions', '-v', multiple=True, default=["3.7", "3.8", "3.9", "3.10"], show_default=True, help="Python version to setup conda environment with.")
def setup(versions: Tuple[str, ...]):
    """
    Setup builder environment with conda.
    """
    versions = tuple(map(_validate_python_version, versions))
    click.secho("Setup builder environment with conda.", fg="green")
    check_conda_installation()
    envs = get_conda_env_set()
    for version in versions:
        env_name = _get_env_name(version)
        if env_name not in envs:
            click.secho(f"Creating {env_name}.", fg="green")
            create_conda_env(env_name, [f"python={version}", "pybind11", "pybind11-stubgen", "numpy", "gym"])
        else:
            click.secho(f"{env_name} already exists. Skipping.", fg="yellow")


@multi_build.command()
@click.option('--versions', '-v', multiple=True, default=["3.7", "3.8", "3.9", "3.10"], show_default=True, help="Python version to remove conda environment with.")
def clean(versions: Tuple[str, ...]):
    """
    Remove builder environment.
    """
    versions = tuple(map(_validate_python_version, versions))
    click.secho("Remove builder environment.", fg="green")
    check_conda_installation()
    envs = get_conda_env_set()
    for version in versions:
        env_name = _get_env_name(version)
        if env_name in envs:
            click.secho(f"Creating {env_name}.", fg="green")
            remove_conda_env(env_name)
        else:
            click.secho(f"{env_name} not exists. Skipping.", fg="yellow")
    click.secho("You may need to delete conda env folders manually.", fg="yellow")


@multi_build.command()
@click.option('--versions', '-v', multiple=True, default=["3.7", "3.8", "3.9", "3.10"], show_default=True, help="Python version to build project with.")
@click.option('--aggregate-output', '-a', default=None, type=click.Path(file_okay=False, resolve_path=True, path_type=Path), help="Folder to aggregate build results into. [default: {workdir}/build/slxpy{plat}]")
@click.option('--aggregate/--no-aggregate', default=True, show_default=True, help="Aggregate build results or not.")
@click.pass_context
def run(ctx: click.Context, versions: Tuple[str, ...], aggregate_output: Path, aggregate: bool):
    """
    Build project for multiple Python versions.
    """
    workdir: Path = ctx.obj["workdir"]
    ensure_slxpy_project(workdir)
    versions = tuple(map(_validate_python_version, versions))
    click.secho("Setup builder environment with conda.", fg="green")
    check_conda_installation()
    envs = get_conda_env_set()
    target_envs = set(map(_get_env_name, versions))
    if not target_envs.issubset(envs):
        raise click.BadParameter(f"Missing environments: {target_envs - envs}, please run \"slxpy build --setup-builder\" first.")
    for version in versions:  # Not using target_envs as set is not ordered.
        env_name = _get_env_name(version)
        click.secho(f"Building project in {env_name}.", fg="green")
        run_in_conda_env(env_name, workdir, ["python", "setup.py", "build_ext"])

    if aggregate:
        if aggregate_output is None:
            aggregate_output = workdir / "build" / f"slxpy.{get_plat_name()}"
        aggregate_output = aggregate_output.absolute()
        aggregate_output.mkdir(exist_ok=True, parents=False)

        click.secho(f"Aggregate build results to {aggregate_output}.", fg="green")
        for version in versions:
            plat_specifier = get_plat_specifier(version)
            build_dir = workdir / "build" / f"lib{plat_specifier}"
            shutil.copytree(build_dir, aggregate_output, dirs_exist_ok=True)


def _validate_python_version(version_str: str):
    from packaging.version import parse as parse_version, Version
    version = parse_version(version_str)
    if version.major != 3:
        raise click.BadParameter(f"Only Python 3.x is supported.")
    elif version.minor < 7:
        raise click.BadParameter(f"Only Python 3.7 and above is supported.")
    else:
        major_minor: Version = parse_version(f"{version.major}.{version.minor}")
        if version != major_minor:
            click.secho(f"Warning: micro will be ignored, {version.base_version} -> {major_minor}", fg="yellow")
        return str(major_minor)


def _get_env_name(version: str):
    return f"slxpy-builder-{version}"



from pathlib import Path
from typing import List

import click

import slxpy.common.constants as C
from slxpy.cli.utils import ensure_slxpy_project


@click.command()
@click.option('--build/--no-build', default=False, help="With build folder.")
@click.option('--asset/--no-asset', default=True, help="Without slxpy generated assets.")
@click.option('--model/--no-model', default=True, help="Without Simulink mdoel sources.")
@click.option('--out', default=None, type=click.Path(dir_okay=False, resolve_path=True, path_type=Path), help="Output zip file name.")
@click.pass_context
def pack(ctx: click.Context, build: bool, asset: bool, model: bool, out: Path):
    """
    Pack assets for transfer.
    """
    workdir: Path = ctx.obj["workdir"]
    ensure_slxpy_project(workdir)

    if out is None: out = Path(f"{workdir.name}.zip")
    parentdir = workdir.parent
    import zipfile, glob

    def writedir(z: zipfile.ZipFile, dir: Path):
        assert dir.exists(), f"Folder not exist: {dir.relative_to(workdir)}"
        assert dir.is_dir()
        for p in dir.glob("**/*"):
            if p.is_dir(): continue
            z.write(p, arcname=p.relative_to(parentdir))
        click.echo(f"Add folder: {dir.relative_to(workdir)}")
    def writefile(z: zipfile.ZipFile, file: Path):
        assert file.exists(), f"File not exist: {file.relative_to(workdir)}"
        assert file.is_file()
        z.write(file, arcname=file.relative_to(parentdir))
        click.echo(f"Add file:   {file.relative_to(workdir)}")

    try:
        with zipfile.ZipFile(out, mode="w", compression=zipfile.ZIP_LZMA) as z:
            writefile(z, workdir / C.model_config_name)
            writefile(z, workdir / C.env_config_name)
            if model:
                writedir(z, workdir / C.model_dir)
            if asset:
                from slxpy.backend.renderer import assets
                writedir(z, workdir / "include")
                generated_files = [asset.name for asset in assets] + [C.project_ir_name]
                for f in generated_files:
                    writefile(z, workdir / f)
            if build:
                writedir(z, workdir / "build")
    except:
        out.unlink()
        raise



from dataclasses import dataclass, field, fields
from typing import ClassVar, Dict, List, Optional, TYPE_CHECKING, TextIO, Tuple, Union

import rtoml
import numpy as np

import slxpy.common.constants as C
from slxpy.common.formatter import format_float
from slxpy.common.mapping import dtype_mapping
from slxpy.common.init_config import InitConfig
from slxpy.common.space_config import SpaceConfig

if TYPE_CHECKING:
    from slxpy.common.context import Module

@dataclass
class GymConfig:
    action_key: Optional[str]
    observation_key: Optional[str]
    reward_key: Optional[str]
    done_key: Optional[str]
    info: Union[bool, List[str]]

    action_space: SpaceConfig
    observation_space: SpaceConfig
    reward_range: Tuple[Union[float, str], Union[float, str]]
    type_coercion: bool

    @staticmethod
    def reconstruct(d: dict):
        reward_range = tuple(d["reward_range"])
        assert len(reward_range) == 2 and all(isinstance(r, (float, str)) for r in reward_range), \
            "reward_range must be a tuple of two floats or +/-inf, not ints"
        return GymConfig(
            action_key=d.get("action_key", None),
            observation_key=d.get("observation_key", None),
            reward_key=d.get("reward_key", None),
            done_key=d.get("done_key", None),
            info=d["info"],
            action_space=SpaceConfig.reconstruct(d["action_space"]),
            observation_space=SpaceConfig.reconstruct(d["observation_space"]),
            reward_range=reward_range,
            type_coercion=d.get("type_coercion", False)
        )

    def asdict(self, dict_filter):
        d = {
            "action_key": self.action_key,
            "observation_key": self.observation_key,
            "reward_key": self.reward_key,
            "done_key": self.done_key,
            "info": self.info,
            "action_space": self.action_space.asdict(dict_filter),
            "observation_space": self.observation_space.asdict(dict_filter),
            "reward_range": self.reward_range,
            "type_coercion": self.type_coercion
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @property
    def reward_initializer(self):
        dtype = np.dtype(np.float64)
        return f"{format_float(self.reward_range[0], dtype)}, {format_float(self.reward_range[1], dtype)}"

    @property
    def unique_boxes(self):
        s = set()
        if self.action_space.type == "Box":
            s.add(self.action_space.dtype)
        if self.observation_space.type == "Box":
            s.add(self.observation_space.dtype)
        return [dtype_mapping[x.name] for x in s]

    @staticmethod
    def default():
        return GymConfig(None, None, None, None, True, SpaceConfig.default(), SpaceConfig.default(), ("-inf", "inf"), True)

@dataclass
class ResetConfig:
    first_step: bool
    input: Optional[str]
    output: Optional[str]

    @staticmethod
    def reconstruct(d: dict):
        first_step = d["first_step"]
        input = d.get("input", None)
        output = d.get("output", None)
        # Consider warn on two conditions below
        if first_step and output is not None:
            output = None
        elif not first_step and input is not None:
            input = None
        return ResetConfig(
            first_step=d["first_step"],
            input=input,
            output=output
        )

    def asdict(self, dict_filter):
        d = {
            "first_step": self.first_step,
            "input": self.input,
            "output": self.output,
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @staticmethod
    def default():
        return ResetConfig(True, None, None)

@dataclass
class Config:
    use_raw: bool
    use_gym: bool
    use_rng: bool
    use_vec: bool
    vec_parallel: bool
    gym: Optional[GymConfig]
    reset: Optional[ResetConfig]
    parameter: Dict[str, InitConfig]

    VERSION: ClassVar[str] = C.metadata_version

    @staticmethod
    def reconstruct(d: dict):
        use_gym = d["use_gym"]
        gym_dict = d.get("gym", None)
        # Consider warn on two conditions below
        if use_gym:
            gym_config = GymConfig.default() if gym_dict is None else GymConfig.reconstruct(gym_dict)
        else:
            gym_config = None
        if "reset" in d:
            reset_config = ResetConfig.reconstruct(d["reset"])
        else:
            reset_config = ResetConfig.default()
        parameter_dict = d.get("parameter", {})
        return Config(
            use_raw=d["use_raw"],
            use_gym=use_gym,
            use_rng=d["use_rng"],
            use_vec=d["use_vec"],
            vec_parallel=d["vec_parallel"],
            gym=gym_config,
            reset=reset_config,
            parameter={k: InitConfig.reconstruct(v) for k, v in parameter_dict.items()}
        )

    def asdict(self, dict_filter):
        d = {
            "use_raw": self.use_raw,
            "use_gym": self.use_gym,
            "use_rng": self.use_rng,
            "use_vec": self.use_vec,
            "vec_parallel": self.vec_parallel,
            "gym": self.gym.asdict(dict_filter) if self.use_gym else None,
            "reset": self.reset.asdict(dict_filter),
            "parameter": {k: v.asdict(dict_filter) for k, v in self.parameter.items()},
        }
        assert len(d) == len(fields(self))  # Ensure no left-out
        return dict_filter(d)

    @staticmethod
    def load(fp: TextIO):
        cfg = rtoml.load(fp)
        assert cfg["__version__"] == Config.VERSION
        return Config.reconstruct(cfg)

    @staticmethod
    def default():
        return Config(
            use_raw=True,
            use_gym=True,
            use_rng=True,
            use_vec=True,
            vec_parallel=False,
            gym=GymConfig.default(),
            reset=ResetConfig.default(),
            parameter={}
        )

    def check_basic_compatibility(self, module: "Module"):
        if not self.use_gym and not self.use_raw:
            return

        # NOTE: The following 3 checks are technically unnecessary
        # But they are here to make sure that the user is aware of
        # correctness of environment configuration.
        # No external input -> Uncontrollable
        # No external output -> Unobservable
        # No parameters -> No randomness, totally deterministic
        #   (even with Simulink Random block, as the seed is fixed)
        if "external_inputs" not in module.model_class.type_mapping:
            raise ValueError("Model must have at least one inport.")
        if "external_outputs" not in module.model_class.type_mapping:
            raise ValueError("Model must have at least one outport.")
        if "instance_parameters" not in module.model_class.type_mapping:
            raise ValueError("Model must have at least one parameter.")

        if self.use_gym:
            # Gym compatibility check
            act_type = module.types.lookup(module.model_class.type_mapping["external_inputs"])
            obs_type = module.types.lookup(module.model_class.type_mapping["external_outputs"])

            assert self.gym.action_key is None or self.gym.action_key in act_type.field_dict
            assert len(obs_type.fields) >= 3
            assert self.gym.observation_key is None or self.gym.observation_key in obs_type.field_dict
            assert self.gym.reward_key is None or self.gym.reward_key in obs_type.field_dict
            assert self.gym.done_key is None or self.gym.done_key in obs_type.field_dict

            if isinstance(self.gym.info, list):
                for s in self.gym.info: assert s in obs_type.field_dict

        param_type = module.types.lookup(module.model_class.type_mapping["instance_parameters"])

        for k, p in self.parameter.items():
            if k.startswith("@"):
                assert p.type == "custom"
            if p.type == "custom":
                continue
            if k not in param_type.field_dict:
                raise KeyError(f"Environment init parameter '{k}' not found in {param_type.name}")
            field = param_type.field_dict[k]
            p.check_basic_compatibility(field)

    def expand_config(self, module: "Module"):
        if self.use_gym:
            act_type = module.types.lookup(module.model_class.type_mapping["external_inputs"])
            obs_type = module.types.lookup(module.model_class.type_mapping["external_outputs"])
            if self.gym.action_key is None:
                self.gym.action_key = act_type.fields[0].name
            if self.gym.observation_key is None:
                self.gym.observation_key = obs_type.fields[0].name
            if self.gym.reward_key is None:
                self.gym.reward_key = obs_type.fields[1].name
            if self.gym.done_key is None:
                self.gym.done_key = obs_type.fields[2].name
            if self.gym.info == True:
                used_fields = (self.gym.observation_key, self.gym.reward_key, self.gym.done_key)
                self.gym.info = [field.name for field in obs_type.fields if field.name not in used_fields]
            elif self.gym.info == False:
                self.gym.info = []



import textwrap
from pathlib import Path

import click

import slxpy.common.constants as C
from slxpy.cli.utils import ensure_slxpy_project, get_plat_specifier, is_debug


@click.command()
@click.pass_context
def frontend(ctx: click.Context):
    """
    Execute slxpy frontend.
    Transform simulink metadata and user config into IR.
    """
    workdir: Path = ctx.obj["workdir"]
    ensure_slxpy_project(workdir)

    click.echo("Execute frontend.", nl=False)
    from slxpy.frontend.frontend import adapt_metadata
    adapt_metadata(workdir)
    click.secho(" SUCCESS", fg="green")

@click.command()
@click.pass_context
def backend(ctx: click.Context):
    """
    Execute slxpy backend.
    Transform IR into binding code and build scripts.
    """
    workdir: Path = ctx.obj["workdir"]
    ensure_slxpy_project(workdir)

    click.echo("Execute backend.", nl=False)
    from slxpy.backend.renderer import render
    render(workdir, is_debug())
    click.secho(" SUCCESS", fg="green")

@click.command()
@click.option('--build', is_flag=True, help="Also build the project.")
@click.pass_context
def generate(ctx: click.Context, build: bool):
    """
    Execute slxpy frontend and backend.
    Produce a self-contained, portable source project.
    """
    workdir: Path = ctx.obj["workdir"]
    ensure_slxpy_project(workdir)

    ctx.invoke(frontend)
    ctx.invoke(backend)
    if build:
        import subprocess, sys
        args = [sys.executable, "setup.py", "build"]
        click.echo(f"Run \"{' '.join(args)}\" to build extension.")
        cp = subprocess.run(args, cwd=workdir)
        cp.check_returncode()
        plat_specifier = get_plat_specifier()
        libdir = workdir / "build" / f"lib{plat_specifier}"
        click.echo(f"\nBuild successful. Check {libdir} for output.")
    else:
        output_text = textwrap.dedent(f"""\
        To build the extension, run in command line:
            > cd "{workdir}"
            > python setup.py build
        """).strip()
        click.echo(output_text)



import jinja2

from slxpy.common.mapping import simulink_mapping

def add_filters(env: jinja2.Environment):
    env.filters["prefix"] = lambda s, prefix: prefix + s
    def field_desc(s: str):
        for m in simulink_mapping:
            if s.endswith(f"_{m[1]}"):
                return m[2]
        return "Unknown"
    env.filters["field_desc"] = field_desc
    env.filters["select_struct"] = lambda l: [s for s in l if not s.is_enum]
    env.filters["select_enum"] = lambda l: [s for s in l if s.is_enum]



from pathlib import Path

import click


@click.command()
@click.pass_context
def init(ctx: click.Context):
    """
    Initialize slxpy working directory.
    """
    workdir: Path = ctx.obj["workdir"]

    from slxpy.frontend.init import init_interactive
    init_interactive(workdir)



import setuptools
from pathlib import Path

cwd = Path(__file__).parent
package_dir = cwd / "slxpy"
package_data = [
    str(f.relative_to(package_dir)) for f in (package_dir / "include").glob("**/*") if f.is_file()
] + [
    str(f.relative_to(package_dir)) for f in (package_dir / "templates").glob("**/*") if f.is_file()
]
long_description = (cwd / "README.md").read_text()

setuptools.setup(
    name='slxpy',
    version='1.3.0',
    description = "Simulink Python binding generator.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    license = "MIT",
    keywords = ["simulink", "c++", "gym"],
    author="Jiang Yuxuan",
    author_email="jyx21@mails.tsinghua.edu.cn",
    url="https://github.com/jjyyxx/slxpy",
    classifiers = [
        "Development Status :: 4 - Beta",
        "Programming Language :: Python"
    ],

    packages=[
        "slxpy",
        "slxpy.common",
        "slxpy.frontend",
        "slxpy.backend",
        "slxpy.cli",
    ],
    package_data={"slxpy": package_data},

    install_requires=[
        "pybind11",
        "pybind11-stubgen",
        "Jinja2>=3.0",
        "rtoml",
        "importlib_resources",
        "packaging",
        "click>=8.0",
        "numpy"
    ],
    entry_points={
        'console_scripts': [
            'slxpy = slxpy.cli:entry_point',
        ],
    },
)



import rtoml
from dataclasses import dataclass
from typing import ClassVar, Literal, TextIO

import slxpy.common.constants as C

@dataclass
class SimulinkConfig:
    solver: Literal["FixedStepAuto", "FixedStepDiscrete","ode8", "ode5", "ode4", "ode3", "ode2", "ode1", "ode14x"] = "FixedStepAuto"
    absolute_time: bool = False
    integer_code: bool = False
    non_finite: bool = False
    complex: bool = False
    continuous_time: bool = False
    variable_size_signal: bool = False
    non_inlined_sfcn: bool = False

    @staticmethod
    def reconstruct(d: dict):
        return SimulinkConfig(
            solver=d["solver"],
            absolute_time=d["absolute_time"],
            integer_code=d["integer_code"],
            non_finite=d["non_finite"],
            complex=d["complex"],
            continuous_time=d["continuous_time"],
            variable_size_signal=d["variable_size_signal"],
            non_inlined_sfcn=d["non_inlined_sfcn"]
        )

@dataclass
class CppConfig:
    class_name: str
    namespace: str

    @staticmethod
    def reconstruct(d: dict):
        return CppConfig(
            class_name=d["class_name"],
            namespace=d["namespace"]
        )

@dataclass
class InfoConfig:
    AUTO: ClassVar[str] = "<auto>"
    NO_LICENSE: ClassVar[str] = "<no-license>"
    description: str = AUTO
    version: str = AUTO
    author: str = AUTO
    license: str = NO_LICENSE

    @staticmethod
    def reconstruct(d: dict):
        return InfoConfig(
            description=d["description"],
            version=d["version"],
            author=d["author"],
            license=d["license"]
        )

@dataclass
class Config:
    model: str
    simulink: SimulinkConfig
    cpp: CppConfig
    info: InfoConfig
    VERSION: ClassVar[str] = C.metadata_version

    @staticmethod
    def reconstruct(d: dict):
        return Config(
            model=d["model"],
            simulink=SimulinkConfig.reconstruct(d["simulink"]),
            cpp=CppConfig.reconstruct(d["cpp"]),
            info=InfoConfig.reconstruct(d["info"])
        )

    @staticmethod
    def load(fp: TextIO):
        cfg = rtoml.load(fp)
        assert cfg["__version__"] == Config.VERSION
        return Config.reconstruct(cfg)

    @staticmethod
    def default(model: str, class_name: str, namespace: str):
        return Config(
            model=model,
            simulink=SimulinkConfig(),
            cpp=CppConfig(class_name=class_name, namespace=namespace),
            info=InfoConfig()
        )



import json, glob, os
from pathlib import Path

from slxpy.common.context import Context, FieldMode, Module, ModelClass, Type, Field, Method, TypeContainer, Enumerator
from slxpy.common.model_config import Config as ModelConfig, InfoConfig
from slxpy.common.env_config import Config as EnvConfig
import slxpy.common.constants as C

def adapt_metadata(workdir: Path):
    modeldir = workdir / C.model_dir

    with (workdir / C.model_config_name).open() as f: model_config = ModelConfig.load(f)
    with (workdir / C.env_config_name).open() as f: env_config = EnvConfig.load(f)

    metadata = json.loads((workdir / C.metadata_name).read_text())
    assert metadata["__version__"] == Context.VERSION, "Metadata version incompatible"

    # TODO: Also check metadata["features"]
    name, model_class, structs = metadata["name"], metadata["model_class"], metadata["structs"]
    doc = metadata["description"] if model_config.info.description == InfoConfig.AUTO else model_config.info.description
    author = metadata["author"] if model_config.info.author == InfoConfig.AUTO else model_config.info.author
    version = metadata["version"] if model_config.info.version == InfoConfig.AUTO else model_config.info.version
    license = "" if model_config.info.license == InfoConfig.NO_LICENSE else model_config.info.license

    model_class_ctx = ModelClass(
        name=model_class["name"], doc=doc, namespace=model_class["namespace"],
        identifier=model_class["identifier"], sample_time=metadata["sample_time"],
        methods=[Method(name=me["name"], doc="") for me in model_class["methods"]],
        fields=[make_field(fi, True) for fi in model_class["fields"]],
        field_mapping=model_class["field_mapping"],
        type_mapping=model_class["type_mapping"],
    )
    types = TypeContainer([make_type(st) for st in structs], model_class_ctx)
    module = Module(
        name=name, doc=doc, version=version, author=author, license=license,
        model_class=model_class_ctx, types=types, env=env_config
    )
    module.env.check_basic_compatibility(module)
    module.env.expand_config(module)
    context = Context(
        sources=[
            # For nested model source in Simscape models
            os.path.relpath(x, modeldir).replace("\\","/") for x in (
                glob.glob(str(modeldir / "*.cpp")) +
                glob.glob(str(modeldir / "*.c")) +
                glob.glob(str(modeldir / '*' / "*.c"))
            )
        ],
        headers=[f"{name}.h"],
        module=module
    )
    with (workdir / C.project_ir_name).open("w") as f: context.dump(f, debug=False)

def make_type(st: dict):
    if st.get("is_enum", False):
        return Type(
            name=st["name"], doc="",
            location=st["location"],
            is_enum=True,
            fields=[make_enumerator(fi) for fi in st["fields"]]
        )
    else:
        return Type(
            name=st["name"], doc="",
            location=st["location"],
            is_enum=False,
            fields=[make_field(fi, False) for fi in st["fields"]]
        )

# Consider using Flag
field_mode_mapping = {
    ("model", False): FieldMode.MODEL,
    ("plain", False): FieldMode.PLAIN,
    ("plain", True): FieldMode.PLAIN_ARRAY,
    ("struct", False): FieldMode.STRUCT,
    ("struct", True): FieldMode.STRUCT_ARRAY,
    ("enum", False): FieldMode.ENUM,
    ("enum", True): FieldMode.ENUM_ARRAY,
    ("pointer", False): FieldMode.POINTER,
    ("pointer", True): FieldMode.POINTER_ARRAY,
}
def ensure_list(list_like):
    return list_like if isinstance(list_like, list) else [list_like]
def make_field(fi: dict, model: bool):
    mode = "model" if model else fi["mode"]
    is_array = fi.get("is_array", False)
    if mode == "std":
        raise NotImplementedError("No std class support yet.")
    if mode == "struct" or mode == "enum":
        type = fi["type"]
    elif mode == "pointer":
        type = fi["underlying"]
    else:
        type = None
    return Field(
        name=fi["name"], doc="",
        mode=field_mode_mapping[(mode, is_array)],
        raw_shape=ensure_list(fi["shape"]) if is_array else None,
        type=type,
    )

def make_enumerator(fi: dict):
    return Enumerator(
        name=fi["name"], doc="",
        value=fi["value"]
    )



# IMPORTANT: Ensure same with slxpy matlab package

model_config_name = "model.toml"
env_config_name = "env.toml"
metadata_version = "1.0.0"
metadata_name = "metadata.json"
model_dir = "model"
project_ir_name = "ir.json"
